@@ -11,6 +11,7 @@ import pytest
 import responses
 import requests
 import jwt
+from urllib3.exceptions import ConnectTimeoutError, MaxRetryError
 from ibm_cloud_sdk_core import BaseService, DetailedResponse
 from ibm_cloud_sdk_core import ApiException
 from ibm_cloud_sdk_core import CP4DTokenManager
@@ -22,12 +23,13 @@ from ibm_cloud_sdk_core.utils import strip_extra_slashes
 
 class IncludeExternalConfigService(BaseService):
     default_service_url = 'https://servicesthatincludeexternalconfig.com/api'
+
     def __init__(
-            self,
-            api_version: str,
-            authenticator: Optional[Authenticator] = None,
-            trace_id: Optional[str] = None
-        ) -> None:
+        self,
+        api_version: str,
+        authenticator: Optional[Authenticator] = None,
+        trace_id: Optional[str] = None
+    ) -> None:
         BaseService.__init__(
             self,
             service_url=self.default_service_url,
@@ -38,16 +40,17 @@ class IncludeExternalConfigService(BaseService):
         self.trace_id = trace_id
         self.configure_service('include-external-config')
 
+
 class AnyServiceV1(BaseService):
     default_url = 'https://gateway.watsonplatform.net/test/api'
 
     def __init__(
-            self,
-            version: str,
-            service_url: str = default_url,
-            authenticator: Optional[Authenticator] = None,
-            disable_ssl_verification: bool = False
-        ) -> None:
+        self,
+        version: str,
+        service_url: str = default_url,
+        authenticator: Optional[Authenticator] = None,
+        disable_ssl_verification: bool = False
+    ) -> None:
         BaseService.__init__(
             self,
             service_url=service_url,
@@ -137,6 +140,9 @@ def test_url_encoding():
         }),
         content_type='application/json')
 
+    # Set Host as a default header on the service.
+    service.set_default_headers({'Host': 'alternatehost.ibm.com:443'})
+
     response = service.op_with_path_params(path0, path1)
 
     assert response is not None
@@ -144,6 +150,10 @@ def test_url_encoding():
     assert path_encoded in responses.calls[0].request.url
     assert 'version=2017-07-07' in responses.calls[0].request.url
 
+    # Verify that the Host header was set in the request.
+    assert responses.calls[0].request.headers.get(
+        'Host') == 'alternatehost.ibm.com:443'
+
 
 @responses.activate
 def test_stream_json_response():
@@ -152,7 +162,8 @@ def test_stream_json_response():
     path = '/v1/streamjson'
     test_url = service.default_url + path
 
-    expected_response = json.dumps({"id": 1, "rev": "v1", "content": "this is a document"})
+    expected_response = json.dumps(
+        {"id": 1, "rev": "v1", "content": "this is a document"})
 
     # print("Expected response: ", expected_response)
 
@@ -206,6 +217,7 @@ def test_fail_http_config():
     with pytest.raises(TypeError):
         service.with_http_config(None)
 
+
 @responses.activate
 def test_cwd():
     file_path = os.path.join(
@@ -230,6 +242,7 @@ def test_cwd():
     assert service.service_url == 'https://cwdserviceurl'
     assert service.authenticator is not None
 
+
 @responses.activate
 def test_iam():
     file_path = os.path.join(
@@ -263,6 +276,7 @@ def test_iam():
     service.any_service_call()
     assert "grant-type%3Aapikey" in responses.calls[0].request.body
 
+
 def test_no_auth():
     class MadeUp:
         def __init__(self):
@@ -292,12 +306,13 @@ def test_for_cp4d():
     assert service.authenticator.token_manager is not None
     assert service.authenticator.token_manager.username == 'my_username'
     assert service.authenticator.token_manager.password == 'my_password'
-    assert service.authenticator.token_manager.url == 'my_url/v1/preauth/validateAuth'
+    assert service.authenticator.token_manager.url == 'my_url/v1/authorize'
     assert isinstance(service.authenticator.token_manager, CP4DTokenManager)
 
 
 def test_disable_ssl_verification():
-    service1 = AnyServiceV1('2017-07-07', authenticator=NoAuthAuthenticator(), disable_ssl_verification=True)
+    service1 = AnyServiceV1(
+        '2017-07-07', authenticator=NoAuthAuthenticator(), disable_ssl_verification=True)
     assert service1.disable_ssl_verification is True
 
     service1.set_disable_ssl_verification(False)
@@ -342,7 +357,8 @@ def test_response_with_no_body():
 
 def test_has_bad_first_or_last_char():
     with pytest.raises(ValueError) as err:
-        basic_authenticator = BasicAuthenticator('{my_username}', 'my_password')
+        basic_authenticator = BasicAuthenticator(
+            '{my_username}', 'my_password')
         AnyServiceV1('2018-11-20', authenticator=basic_authenticator).prepare_request(
             responses.GET,
             'https://gateway.watsonplatform.net/test/api'
@@ -352,6 +368,7 @@ def test_has_bad_first_or_last_char():
     ) == 'The username and password shouldn\'t start or end with curly brackets or quotes. '\
          'Please remove any surrounding {, }, or \" characters.'
 
+
 @responses.activate
 def test_request_server_error():
     responses.add(
@@ -369,6 +386,7 @@ def test_request_server_error():
     except ApiException as err:
         assert err.message == 'internal server error'
 
+
 @responses.activate
 def test_request_success_json():
     responses.add(
@@ -384,13 +402,15 @@ def test_request_success_json():
     detailed_response = service.send(prepped)
     assert detailed_response.get_result() == {'foo': 'bar'}
 
-    service = AnyServiceV1('2018-11-20', authenticator=BasicAuthenticator('my_username', 'my_password'))
+    service = AnyServiceV1(
+        '2018-11-20', authenticator=BasicAuthenticator('my_username', 'my_password'))
     service.set_default_headers({'test': 'header'})
     service.set_disable_ssl_verification(True)
     prepped = service.prepare_request('GET', url='')
     detailed_response = service.send(prepped)
     assert detailed_response.get_result() == {'foo': 'bar'}
 
+
 @responses.activate
 def test_request_success_response():
     responses.add(
@@ -406,6 +426,7 @@ def test_request_success_response():
     detailed_response = service.send(prepped)
     assert detailed_response.get_result() == {"foo": "bar"}
 
+
 @responses.activate
 def test_request_fail_401():
     responses.add(
@@ -423,6 +444,7 @@ def test_request_fail_401():
     except ApiException as err:
         assert err.message == 'Unauthorized: Access is denied due to invalid credentials'
 
+
 def test_misc_methods():
 
     class MockModel:
@@ -456,6 +478,7 @@ def test_misc_methods():
     res_str = service._convert_list(temp)
     assert res_str == 'default,123'
 
+
 def test_default_headers():
     service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
     service.set_default_headers({'xxx': 'yyy'})
@@ -463,6 +486,7 @@ def test_default_headers():
     with pytest.raises(TypeError):
         service.set_default_headers('xxx')
 
+
 def test_set_service_url():
     service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
     with pytest.raises(ValueError) as err:
@@ -472,11 +496,13 @@ def test_set_service_url():
 
     service.set_service_url('my_url')
 
+
 def test_http_client():
     auth = BasicAuthenticator('my_username', 'my_password')
     service = AnyServiceV1('2018-11-20', authenticator=auth)
     assert isinstance(service.get_http_client(), requests.sessions.Session)
-    assert service.get_http_client().headers.get('Accept-Encoding') == 'gzip, deflate'
+    assert service.get_http_client().headers.get(
+        'Accept-Encoding') == 'gzip, deflate'
 
     new_http_client = requests.Session()
     new_http_client.headers.update({'Accept-Encoding': 'gzip'})
@@ -486,23 +512,27 @@ def test_http_client():
     with pytest.raises(TypeError):
         service.set_http_client("bad_argument_type")
 
+
 def test_get_authenticator():
     auth = BasicAuthenticator('my_username', 'my_password')
     service = AnyServiceV1('2018-11-20', authenticator=auth)
     assert service.get_authenticator() is not None
 
+
 def test_gzip_compression():
     # Should return uncompressed data when gzip is off
     service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
     assert not service.get_enable_gzip_compression()
-    prepped = service.prepare_request('GET', url='', data=json.dumps({"foo": "bar"}))
+    prepped = service.prepare_request(
+        'GET', url='', data=json.dumps({"foo": "bar"}))
     assert prepped['data'] == b'{"foo": "bar"}'
     assert prepped['headers'].get('content-encoding') != 'gzip'
 
     # Should return compressed data when gzip is on
     service.set_enable_gzip_compression(True)
     assert service.get_enable_gzip_compression()
-    prepped = service.prepare_request('GET', url='', data=json.dumps({"foo": "bar"}))
+    prepped = service.prepare_request(
+        'GET', url='', data=json.dumps({"foo": "bar"}))
     assert prepped['data'] == gzip.compress(b'{"foo": "bar"}')
     assert prepped['headers'].get('content-encoding') == 'gzip'
 
@@ -541,18 +571,96 @@ def test_gzip_compression():
     assert prepped['data'] == b'{"foo": "bar"}'
     assert prepped['headers'].get('content-encoding') == 'gzip'
 
+
 def test_gzip_compression_external():
     # Should set gzip compression from external config
     file_path = os.path.join(
         os.path.dirname(__file__), '../resources/ibm-credentials-gzip.env')
     os.environ['IBM_CREDENTIALS_FILE'] = file_path
-    service = IncludeExternalConfigService('v1', authenticator=NoAuthAuthenticator())
+    service = IncludeExternalConfigService(
+        'v1', authenticator=NoAuthAuthenticator())
     assert service.service_url == 'https://mockurl'
     assert service.get_enable_gzip_compression() is True
-    prepped = service.prepare_request('GET', url='', data=json.dumps({"foo": "bar"}))
+    prepped = service.prepare_request(
+        'GET', url='', data=json.dumps({"foo": "bar"}))
     assert prepped['data'] == gzip.compress(b'{"foo": "bar"}')
     assert prepped['headers'].get('content-encoding') == 'gzip'
 
+
+def test_retry_config_default():
+    service = BaseService(service_url='https://mockurl/',
+                          authenticator=NoAuthAuthenticator())
+    service.enable_retries()
+    assert service.retry_config.total == 4
+    assert service.retry_config.backoff_factor == 0.1
+    assert service.http_client.get_adapter('https://').max_retries.total == 4
+
+    # Ensure retries fail after 4 retries
+    error = ConnectTimeoutError()
+    retry = service.http_client.get_adapter('https://').max_retries
+    retry = retry.increment(error=error)
+    retry = retry.increment(error=error)
+    retry = retry.increment(error=error)
+    retry = retry.increment(error=error)
+    with pytest.raises(MaxRetryError) as retry_err:
+        retry.increment(error=error)
+    assert retry_err.value.reason == error
+
+
+def test_retry_config_disable():
+    # Test disabling retries
+    service = BaseService(service_url='https://mockurl/',
+                          authenticator=NoAuthAuthenticator())
+    service.enable_retries()
+    service.disable_retries()
+    assert service.retry_config is None
+    assert service.http_client.get_adapter('https://').max_retries.total == 0
+
+    # Ensure retries are not started after one connection attempt
+    error = ConnectTimeoutError()
+    retry = service.http_client.get_adapter('https://').max_retries
+    with pytest.raises(MaxRetryError) as retry_err:
+        retry.increment(error=error)
+    assert retry_err.value.reason == error
+
+
+def test_retry_config_non_default():
+    service = BaseService(service_url='https://mockurl/',
+                          authenticator=NoAuthAuthenticator())
+    service.enable_retries(2, 0.3)
+    assert service.retry_config.total == 2
+    assert service.retry_config.backoff_factor == 0.3
+
+    # Ensure retries fail after 2 retries
+    error = ConnectTimeoutError()
+    retry = service.http_client.get_adapter('https://').max_retries
+    retry = retry.increment(error=error)
+    retry = retry.increment(error=error)
+    with pytest.raises(MaxRetryError) as retry_err:
+        retry.increment(error=error)
+    assert retry_err.value.reason == error
+
+
+def test_retry_config_external():
+    file_path = os.path.join(
+        os.path.dirname(__file__), '../resources/ibm-credentials-retry.env')
+    os.environ['IBM_CREDENTIALS_FILE'] = file_path
+    service = IncludeExternalConfigService(
+        'v1', authenticator=NoAuthAuthenticator())
+    assert service.retry_config.total == 3
+    assert service.retry_config.backoff_factor == 0.2
+
+    # Ensure retries fail after 3 retries
+    error = ConnectTimeoutError()
+    retry = service.http_client.get_adapter('https://').max_retries
+    retry = retry.increment(error=error)
+    retry = retry.increment(error=error)
+    retry = retry.increment(error=error)
+    with pytest.raises(MaxRetryError) as retry_err:
+        retry.increment(error=error)
+    assert retry_err.value.reason == error
+
+
 @responses.activate
 def test_user_agent_header():
     service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
@@ -577,16 +685,42 @@ def test_user_agent_header():
     assert response.get_result().request.headers.__getitem__(
         'user-agent') == user_agent_header['User-Agent']
 
+@responses.activate
+def test_reserved_keys(caplog):
+    service = AnyServiceV1('2021-07-02', authenticator=NoAuthAuthenticator())
+    responses.add(
+        responses.GET,
+        'https://gateway.watsonplatform.net/test/api',
+        status=200,
+        body='some text')
+    prepped = service.prepare_request('GET', url='', headers={'key': 'OK'})
+    response = service.send(
+        prepped,
+        headers={'key': 'bad'},
+        method='POST',
+        url='localhost',
+        cookies=None,
+        hooks={'response': []})
+    assert response.get_result().request.headers.__getitem__('key') == 'OK'
+    assert response.get_result().request.url == 'https://gateway.watsonplatform.net/test/api'
+    assert response.get_result().request.method == 'GET'
+    assert response.get_result().request.hooks == {'response': []}
+    assert caplog.record_tuples[0][2] == '"method" has been removed from the request'
+    assert caplog.record_tuples[1][2] == '"url" has been removed from the request'
+    assert caplog.record_tuples[2][2] == '"headers" has been removed from the request'
+    assert caplog.record_tuples[3][2] == '"cookies" has been removed from the request'
+
 def test_files_dict():
     service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
 
     form_data = {}
-    file = open(
+    with open(
         os.path.join(
-            os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r')
-    form_data['file1'] = (None, file, 'application/octet-stream')
+            os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r') as file:
+        form_data['file1'] = (None, file, 'application/octet-stream')
     form_data['string1'] = (None, 'hello', 'text/plain')
-    request = service.prepare_request('GET', url='', headers={'X-opt-out': True}, files=form_data)
+    request = service.prepare_request(
+        'GET', url='', headers={'X-opt-out': True}, files=form_data)
     files = request['files']
     assert isinstance(files, list)
     assert len(files) == 2
@@ -596,16 +730,18 @@ def test_files_dict():
     string1 = files_dict['string1']
     assert string1[0] is None
 
+
 def test_files_list():
     service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
 
     form_data = []
-    file = open(
+    with open(
         os.path.join(
-            os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r')
-    form_data.append(('file1', (None, file, 'application/octet-stream')))
+            os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r') as file:
+        form_data.append(('file1', (None, file, 'application/octet-stream')))
     form_data.append(('string1', (None, 'hello', 'text/plain')))
-    request = service.prepare_request('GET', url='', headers={'X-opt-out': True}, files=form_data)
+    request = service.prepare_request(
+        'GET', url='', headers={'X-opt-out': True}, files=form_data)
     files = request['files']
     assert isinstance(files, list)
     assert len(files) == 2
@@ -615,23 +751,28 @@ def test_files_list():
     string1 = files_dict['string1']
     assert string1[0] is None
 
+
 def test_files_duplicate_parts():
     service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
 
     form_data = []
-    file = open(
+    with open(
         os.path.join(
-            os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r')
-    form_data.append(('creds_file', (None, file, 'application/octet-stream')))
-    file = open(
+            os.path.dirname(__file__), '../resources/ibm-credentials-iam.env'), 'r') as file:
+        form_data.append(
+            ('creds_file', (None, file, 'application/octet-stream')))
+    with open(
         os.path.join(
-            os.path.dirname(__file__), '../resources/ibm-credentials-basic.env'), 'r')
-    form_data.append(('creds_file', (None, file, 'application/octet-stream')))
-    file = open(
+            os.path.dirname(__file__), '../resources/ibm-credentials-basic.env'), 'r') as file:
+        form_data.append(
+            ('creds_file', (None, file, 'application/octet-stream')))
+    with open(
         os.path.join(
-            os.path.dirname(__file__), '../resources/ibm-credentials-bearer.env'), 'r')
-    form_data.append(('creds_file', (None, file, 'application/octet-stream')))
-    request = service.prepare_request('GET', url='', headers={'X-opt-out': True}, files=form_data)
+            os.path.dirname(__file__), '../resources/ibm-credentials-bearer.env'), 'r') as file:
+        form_data.append(
+            ('creds_file', (None, file, 'application/octet-stream')))
+    request = service.prepare_request(
+        'GET', url='', headers={'X-opt-out': True}, files=form_data)
     files = request['files']
     assert isinstance(files, list)
     assert len(files) == 3
@@ -639,11 +780,14 @@ def test_files_duplicate_parts():
         assert part_name == 'creds_file'
         assert file_tuple[0] is not None
 
+
 def test_json():
     service = AnyServiceV1('2018-11-20', authenticator=NoAuthAuthenticator())
-    req = service.prepare_request('POST', url='', headers={'X-opt-out': True}, data={'hello': 'world'})
+    req = service.prepare_request('POST', url='', headers={
+                                  'X-opt-out': True}, data={'hello': 'world'})
     assert req.get('data') == "{\"hello\": \"world\"}"
 
+
 def test_trailing_slash():
     assert strip_extra_slashes('') == ''
     assert strip_extra_slashes('//') == '/'
@@ -656,7 +800,8 @@ def test_trailing_slash():
     assert strip_extra_slashes('https://host/path//') == 'https://host/path/'
     assert strip_extra_slashes('https://host//path//') == 'https://host//path/'
 
-    service = AnyServiceV1('2018-11-20', service_url='https://host/', authenticator=NoAuthAuthenticator())
+    service = AnyServiceV1(
+        '2018-11-20', service_url='https://host/', authenticator=NoAuthAuthenticator())
     assert service.service_url == 'https://host/'
     service.set_service_url('https://host/')
     assert service.service_url == 'https://host/'
@@ -666,7 +811,8 @@ def test_trailing_slash():
                                   data={'hello': 'world'})
     assert req.get('url') == 'https://host//path/'
 
-    service = AnyServiceV1('2018-11-20', service_url='https://host/', authenticator=NoAuthAuthenticator())
+    service = AnyServiceV1(
+        '2018-11-20', service_url='https://host/', authenticator=NoAuthAuthenticator())
     assert service.service_url == 'https://host/'
     service.set_service_url('https://host/')
     assert service.service_url == 'https://host/'
@@ -679,7 +825,8 @@ def test_trailing_slash():
     service.set_service_url(None)
     assert service.service_url is None
 
-    service = AnyServiceV1('2018-11-20', service_url='/', authenticator=NoAuthAuthenticator())
+    service = AnyServiceV1('2018-11-20', service_url='/',
+                           authenticator=NoAuthAuthenticator())
     assert service.service_url == '/'
     service.set_service_url('/')
     assert service.service_url == '/'
@@ -689,14 +836,17 @@ def test_trailing_slash():
                                   data={'hello': 'world'})
     assert req.get('url') == '/'
 
+
 def test_service_url_not_set():
     service = BaseService(service_url='', authenticator=NoAuthAuthenticator())
     with pytest.raises(ValueError) as err:
         service.prepare_request('POST', url='')
     assert str(err.value) == 'The service_url is required'
 
+
 def test_setting_proxy():
-    service = BaseService(service_url='test', authenticator=IAMAuthenticator('wonder woman'))
+    service = BaseService(service_url='test',
+                          authenticator=IAMAuthenticator('wonder woman'))
     assert service.authenticator is not None
     assert service.authenticator.token_manager.http_config == {}
 
@@ -708,22 +858,27 @@ def test_setting_proxy():
     service.set_http_config(http_config)
     assert service.authenticator.token_manager.http_config == http_config
 
-    service2 = BaseService(service_url='test', authenticator=BasicAuthenticator('marvellous', 'mrs maisel'))
+    service2 = BaseService(service_url='test', authenticator=BasicAuthenticator(
+        'marvellous', 'mrs maisel'))
     service2.set_http_config(http_config)
     assert service2.authenticator is not None
 
+
 def test_configure_service():
     file_path = os.path.join(
         os.path.dirname(__file__), '../resources/ibm-credentials-external.env')
     os.environ['IBM_CREDENTIALS_FILE'] = file_path
-    service = IncludeExternalConfigService('v1', authenticator=NoAuthAuthenticator())
+    service = IncludeExternalConfigService(
+        'v1', authenticator=NoAuthAuthenticator())
     assert service.service_url == 'https://externallyconfigured.com/api'
     assert service.disable_ssl_verification is True
     # The authenticator should not be changed as a result of configure_service()
     assert isinstance(service.get_authenticator(), NoAuthAuthenticator)
 
+
 def test_configure_service_error():
-    service = BaseService(service_url='v1', authenticator=NoAuthAuthenticator())
+    service = BaseService(
+        service_url='v1', authenticator=NoAuthAuthenticator())
     with pytest.raises(ValueError) as err:
         service.configure_service(None)
     assert str(err.value) == 'Service_name must be of type string.'
